/*
 * BitSet.h
 *
 *  Created on: Jun 19, 2013
 *      Author: lacasta
 */

#ifndef BITSET_H_
#define BITSET_H_

#include <iosfwd>

class BitSet
{
    public:
        int byte_reverse(int val, int nbit) const;

    private:
        unsigned char *pb; // Bit string starts here
        int nbits; // Number of bits
        int szSlot; // Internal: size of unsigned int slot
        int nSlot; // number of slots
        int nsTrue; // true number of slots
        int nset; // Number of Bits set
        void cpBits(int, unsigned char *);
        void prt(std::ostream &);

    public:
        friend std::ostream& operator<<(std::ostream&, BitSet&);
        // Constructors (int refers always to no. of bits)
        BitSet(int nn = 128, int szW = 8 * sizeof(unsigned char));
        BitSet(const BitSet&);
        BitSet(int, char *);
        BitSet(int, unsigned char *);

        // Destructor
        ~BitSet();
        // Copy operator
        BitSet& operator =(const BitSet& y);

        // Operators
        void new_size(int, int);
        BitSet * Reset(unsigned char *);
        BitSet * Word(int, unsigned char); // Set Word i to value x
        BitSet * Set(int); // Sets    i-th bit
        BitSet * Set(int val, int start, int width); // Sets to val the byte
        BitSet * SetR(int val, int start, int width)
        {
            return Set(byte_reverse(val, width), start, width);
        }
        // at start with length width
        BitSet * Clr(int); // Clears  i-th bit
        BitSet * operator!();
        BitSet * AllZero();
        int operator[](int) const; // alias for CheckBit

        int Bit(int) const; // Checks  i-th bit
        void BitStr(int, const char *); // Set bits according th 0's and 1's in string

        // interface
        unsigned char Word(int x) const;
        unsigned int nWords() const
        {
            return nSlot;
        }

        // Get length of bit stream
        int nBits() const
        {
            return nbits;
        }

        // get number of bits to 1
        int nSet() const
        {
            return nset;
        }
        unsigned long Byte(int, int) const;

        unsigned long ByteR(int start, int width) const
        {
            return byte_reverse(Byte(start, width), width);
        }
        void Print(int, int);
        void Dump(); // Dump internals

};

#endif /* BITSET_H_ */
