/*
 * CalibrationRun.cc
 *
 *  Created on: May 12, 2010
 *      Author: lacasta
 */

#include <iostream>
#include "DAQmanager.h"
#include "CalibrationRun.h"
#include "Analysis.h"
#include "DataMonitor.h"
#include "ScanConfig.h"
#include <ConfigFactory.h>
#include "utils.h"

CalibrationRun::CalibrationRun(DAQmanager *D)
    : RunManager(D)
{

}

CalibrationRun::~CalibrationRun()
{
}

void CalibrationRun::start_of_run()
{
    ALIBAVA_TRACE_CLASS("");
    ConfigFactory &cfg = ConfigFactory::instance();
    Analysis *ana = dynamic_cast<Analysis *>(cfg.get("analysis"));
    CalibrationConfig *cal = dynamic_cast<CalibrationConfig *>(cfg.get("calibration"));
    ana->update_pedestals(false);

    ScanConfig C = cal->scan_charge() ? cal->charge_scan() : cal->delay_scan();
    DataMonitor *mon = get_daq_manager()->get_monitor();
    if (mon)
        mon->create_signal_profile("injected pulse",
                                   C.npts(),
                                   C.from(),
                                   C.to() );
    get_daq_manager()->set_expected( C.npts()*C.nevts() );
    get_daq_manager()->set_received(0);
}

int CalibrationRun::loop( std::ostream &oheader)
{
    ALIBAVA_TRACE_CLASS("");
    unsigned int rc, i, step;
    ConfigFactory &cfg = ConfigFactory::instance();
    CalibrationConfig *cal = dynamic_cast<CalibrationConfig *>(cfg.get("calibration"));
    ScanConfig C = cal->scan_charge() ? cal->charge_scan() : cal->delay_scan();
    step = C.step();
    oheader << C.npts() << "; "
            << C.from() << "; "
            << C.to()  << "; "
            << C.step();
    if ( ! cal->scan_charge() )
        oheader << "; delay_scan";

    rc = 0;
    unsigned int ntrg;
    Alibava *alibava = get_daq_manager()->get_board();
    alibava->define_scan(C.npts(), C.from(), C.to());
    for (i=0; i<C.npts();i++)
    {
        alibava->reset_trigger_counter();

        if (get_daq_manager()->stop())
            break;

        if ( (rc=acquire(C.from()+ i *step)) )
            return rc;

        if ( (rc=read_data()) )
            return rc;

        ntrg = alibava->get_trigger_counter();
        incNtrig(ntrg);
        alibava->end_of_scan_point(C.from() + i*step, ntrg);

    }
    return 0;
}

int CalibrationRun::acquire(unsigned int val)
{
    ALIBAVA_TRACE_CLASS("");
    int rc = 0;
    ConfigFactory &cfg = ConfigFactory::instance();
    CalibrationConfig *cal = dynamic_cast<CalibrationConfig *>(cfg.get("calibration"));
    ScanConfig C = cal->scan_charge() ? cal->charge_scan() : cal->delay_scan();
    if ( cal->scan_charge() )
        rc = get_daq_manager()->get_board()->calibration(val, cal->delay(), C.nevts());
    else
        rc = get_daq_manager()->get_board()->calibration(cal->charge(), val, C.nevts());

    return rc;
}
