//#define DEBUG 1
#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#else
#warning "NO config"
#endif

#ifndef HAVE_GIO

#define ALIBAVA_SERVER
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <glib.h>
#include <math.h>
#include <signal.h>
#include <getopt.h>
#include <alibava-server-no-gio.h>

#include <iostream>
#include <map>
#include <string>
#include <glib-object.h>
#include "string_utils.h"
#include "Machine.h"
#include "Alibava.h"
static void start_subprocess();
static std::map<int, struct client_info *> clients;
static GMainLoop *loop=NULL;

extern int __is_gui_session__;
TheMachine machine;

static bool do_run=true;

static void quit(int sg)
{
    info_msg("Have to quit....\n");
    g_main_loop_quit(loop);
}

static void no_pipe(int sg)
{
    err_msg ("Broken pipe\n");
    g_main_loop_quit(loop);
}

static int init_daemon()
{
    pid_t pid;
    /**
     Become a deamon.
     */
    if ( (pid=fork())<0)
        return -1;
    else if (pid!=0)
        exit(0); // parent exits

    // Child continues
    setsid(); // become session leader

    chdir("/"); // change working directory
    umask(0); // clear our file mode creation mask

    return 0;
}

int main(int argc, char **argv)
{
    bool is_daemon = true;
    int  fw_version = 2;
    int option_index;
    std::string value;
    static struct option long_options[] =
    {
     { "daemon", 1, 0, 'd'},
     { "firmware-version", 1, 0, 'v'},
     { "help", 0, 0, 'h'},
     {0,0,0,0}
    };
    
    __is_gui_session__ = 0;

    // Initialize
    g_thread_init(NULL);
#if GLIB_CHECK_VERSION(2, 24, 0)
    g_type_init();
#endif
    
    while (true)
    {
        char c;
        option_index = 0;
        c = getopt_long(argc, argv, ":d:v:h", long_options, &option_index);
        if (c == -1 )
            break;
        switch (c)
        {
            case 0:
                break;
            case 'd':
                value = lower_string( optarg );
                if (value != "yes" && value!="no")
                {
                    std::cout << "argument of 'daemon' should be yes or no" << std::endl;
                    return 1;
                }
                if (value=="no")
                    is_daemon = false;

                break;

            case 'v':
                fw_version = atoi(optarg);
                break;
            case 'h':
                std::cout << "alibava-server" << std::endl;
                return 0;
            default:
                std::cout << "Error " << argv[optind-1] << " ";
                if (optind<argc)
                    std::cout << argv[optind];
                std::cout << " [Ughhhh !!!!]" << std::endl;
                return 1;
        }
    }

    g_printerr("alibava-server\n");

    machine.set_fw_version(fw_version);
    if ( is_daemon )
    {
        if (init_daemon())
            return 1;
    }

    openlog("alibavad", LOG_NDELAY|LOG_PID, LOG_USER);
    syslog(LOG_INFO,"Starting the ALIBAVA daemon");

    start_subprocess();

    syslog(LOG_INFO,"Shuting down the ALIBAVA daemon");
    closelog();
    return 0;
}

static int create_socket(const char *name)
{
    char line[128];
    int fd, len;
    struct sockaddr_un unix_addr;

    if ( (fd=socket(AF_UNIX, SOCK_STREAM,0))<0)
    {
        err_msg(_("Could not create socket\n"));
        return -1;
    }

    unlink(name);
    unix_addr.sun_family = AF_UNIX;
    strcpy(unix_addr.sun_path, name);
    len = sizeof(unix_addr.sun_family)+strlen(unix_addr.sun_path);

    // bind the name
    if (bind(fd, (struct sockaddr *)&unix_addr, len)<0)
    {
        sprintf(line, _("Could not bind name %s to socket\n"), unix_addr.sun_path);
        err_msg(line);
        return -2;
    }

    // tell the kernel we are a server
    if (listen(fd, 2)<0)
    {
        sprintf(line, _("Error in listen for socket %s\n"), unix_addr.sun_path);
        err_msg(line);
        return -3;
    }
    
    int flgs = fcntl(fd, F_GETFL) & ~O_NONBLOCK;
    fcntl(fd, F_SETFL, flgs);
    
    return fd;
}

static int accept_connection(int fd, uid_t *uidptr, int *type)
{
    static int counter=0;
    int clifd, len, flgs;
    struct sockaddr_un unix_addr;

    PDEBUG("new connection");

    len = sizeof(unix_addr);
    PDEBUG("...calling accept");
    if ( (clifd=accept(fd, (sockaddr *)&unix_addr, (socklen_t *)&len))<0)
    {
        err_msg(_("Unable to accept connection\n"));
        return -1;
    }
    len -= sizeof(unix_addr.sun_family);
    unix_addr.sun_path[len] = 0;

    if (uidptr!=NULL)
        *uidptr = counter++;

    // get the tipe
    if (type!=NULL)
        *type = *uidptr % 2 ? 0 : 1;

    flgs = fcntl(clifd, F_GETFL) & ~O_NONBLOCK;
    fcntl(clifd, F_SETFL, flgs);
    
    int opt_val = 1;
    setsockopt( clifd,
                IPPROTO_TCP,
                TCP_NODELAY, (char *)&opt_val,
                sizeof(int));
    return clifd;
}

static gboolean server_client_io(GIOChannel * channel, GIOCondition condition, struct client_info *data)
{
    PDEBUG("client IO: %d",condition);

    if (condition == G_IO_HUP)
    {
        printf("file HUP\n");
        return FALSE;
    }

    if (condition == G_IO_OUT)
    {
        PDEBUG("client wants to read");
        exit(1);
    }
    else if (condition == G_IO_IN)
    {
        PDEBUG("client wants to write");
        int rc = machine.process_state(data->io); // data->fd
        if (rc<0)
        {
            // There has been an error, give it back
        }
    }
    return TRUE;
}

static void remove_client(struct client_info *data)
{
    gboolean rc;
    PDEBUG("removing client");

    g_mutex_free( data->mux );
    shutdown(data->fd, SHUT_RDWR);
    shutdown(data->cfd, SHUT_RDWR);
    close(data->fd);
    close(data->cfd);

    do
    {
        rc = g_source_remove_by_user_data(data);
    } while (rc);

    clients.erase(data->uid);
    do_run = false;
}

static gboolean server_client_reads(struct client_info *data)
{
/*
    if (machine.mstate()==TheMachine::ReadingData)
    {
        //        printf("server_client_reads\n");
        if (machine.send_data_block(data->fd, data->size)<0)
        {
            machine.set_to_wait(data->fd);
        }
    }
    */
    return FALSE;
}

static gboolean server_ioctl(GIOChannel * channel, GIOCondition condition, struct client_info *data)
{
    char type;
    int nr, dir, sz, rc = 0;
    int die = 0;
    size_t size;

    if (condition == G_IO_HUP)
    {
        printf("control HUP\n");
        exit(1);
    }

    // first read the command
    read(data->cfd, &data->command, sizeof(int));
    type = _IOC_TYPE (data->command);
    nr = _IOC_NR (data->command);
    dir = _IOC_DIR (data->command);
    sz = _IOC_SIZE (data->command);

    switch (nr)
    {
    case RESET_NR:
        PDEBUG ("I have to RESET");
        break;

    case DIE_NR:
        PDEBUG ("I have to quit");
        die = 1;
        break;

    case ERROR_NR:
        rc = machine.error();
        break;

    case DO_READ_NR:
        read(data->cfd, &size, sz);
        PDEBUG ("do_read size=%d", size);
        data->size = size;
        g_idle_add((GSourceFunc) server_client_reads, data);
        break;
    default:
        PDEBUG ("ioctl Unknown command %x", data->command);
        rc = -1;
        break;
    }
    // write back the return code
    PDEBUG ("ioctl sends %d as rc", rc);
    write(data->cfd, &rc, sizeof(int));
    if (die)
    {
        // we do not wait for a return code when closing
        PDEBUG ("removing clients");
        remove_client(data);
        do_run = false;
    }

    return TRUE;
}

static gpointer server_accepts(gpointer ptr)
{
    int clifd;
    long fd = (long)ptr; 
    int type;
    uid_t uid, duid;
    struct client_info *data;
    unsigned long word = 0xbebecafe;
    GIOChannel *chan;
    GError *error=0;


    while(do_run)
    {
        PDEBUG("new connection");

        clifd = accept_connection(fd, &uid, &type);
        if (!do_run)
        {
            PDEBUG("have to quit");
            break;
        }
        PDEBUG("...accepted: %d", clifd);
        // Check if we have this client in the list
        duid = 2*(uid/2);

        std::map<int, struct client_info *>::iterator ip = clients.find(duid);
        if (ip==clients.end())
        {
            PDEBUG("...new client");
            data = g_new0(struct client_info,1);
            data->uid = duid;
            data->mux = g_mutex_new();
            clients[duid]=data;
        }
        else
        {
            PDEBUG("...existing client");
            data = ip->second;
        }

        if (type)
        { // this is the ioctl channel
            PDEBUG("...creating control channel");
            data->cfd=clifd;
#ifdef G_PLATFORM_WIN32
            chan = data->control = g_io_channel_win32_new_socket (clifd);
#else
            chan = data->control = g_io_channel_unix_new(clifd);
#endif
            g_io_add_watch(data->control, (GIOCondition) (G_IO_IN | G_IO_PRI | G_IO_HUP), (GIOFunc) server_ioctl, data);
            PDEBUG("...done");
        }
        else
        {
            PDEBUG("...creating data channel\n");
            data->fd = clifd;
#ifdef G_PLATFORM_WIN32
            chan = data->io = g_io_channel_win32_new_socket (clifd);
#else
            chan = data->io = g_io_channel_unix_new(clifd);
#endif
            g_io_add_watch(data->io, (GIOCondition) (G_IO_IN | G_IO_PRI | G_IO_HUP), (GIOFunc) server_client_io, data);
            PDEBUG("...done");
        }

        g_io_channel_set_encoding(chan, 0, &error);
        g_io_channel_set_buffered(chan, FALSE);
        g_io_channel_ref(chan);

        PDEBUG("Will send handshake");
        int rc = write(clifd, &word, sizeof(unsigned long));
        PDEBUG("Send handshake %x. rc=%d", word, rc);
        if ( rc < 0 )
        {
            err_msg(_("Cannot send handshake word: %s\n"), strerror(errno));
        }
    }

    return NULL;
}

/*
 Starts the daemon process. It creates a socket that will communicate
 the user application with the server
 */
static void start_subprocess()
{
    int sfd;
    FILE *pf;

    
    PDEBUG("Server starting");
    
    pf = fopen("/tmp/alibava.lock", "w");
    fprintf(pf, "%d", getpid());
    fclose(pf);

    PDEBUG("Creating socket");
    sfd= create_socket("/tmp/alibava.socket");
    if (sfd<0)
    	exit(1);
    PDEBUG("...socket fildes: %d", sfd);


    // init the random number generator
    srand(time(0));


    /** TODO: Make things easy (for windows), since we will not have
     *  multiple users... 
     * Just create an extra thread which calls accept (we limit listen 
     * to 2 connections) andd the watchers to the created streams.
     *  
     */
    // init the list of clients
    clients.clear();
    loop = g_main_loop_new(NULL, FALSE);
    g_thread_create(server_accepts,(gpointer)sfd, FALSE,NULL);

    signal(SIGPIPE, no_pipe);
    signal(SIGINT, quit);
    signal(SIGQUIT, quit);
    signal(SIGTERM, quit);

    PDEBUG("Start loop");
    g_main_loop_run(loop);
    close(sfd);
    unlink("/tmp/alibava.lock");
    unlink("/tmp/alibava.socket");
    return;
}

#endif
